/*
** Module   : SocketImpl.java
** Abstract : A class for implementing the SOCKET method and attributes
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130204 Created initial version.
** 002 CS  20130204 Moved the SSL-SERVER attribute to it's own interface.
** 003 CA  20130221 Added Deletable support. setType was renamed to setResourceType.
** 004 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CS  20130404 Implement runtime support
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import javax.net.ssl.*;

/**
 * A class for implementing the SOCKET method and attributes
 */
public class SocketImpl
extends SensitiveResource
implements com.goldencode.p2j.util.Socket,
           Sensitive,
           Connectable,
           SslServerAttribute
{
   /**  The socket through which the communication is done*/
   private java.net.Socket sock;
   
   
   /** The address to which to connect*/
   private InetSocketAddress address;
   
   /** output socket stream*/
   private BufferedOutputStream out;
   
   /** input socket stream */
   private BufferedInputStream in;
   
   /** bytes read*/
   private int bytesRead = 0;
   
   /** bytes written*/
   private int bytesWritten = 0;

   /** connected*/
   private boolean connected;
   
   /**
    * Default constructor which initializes SOCKET specific settings.
    */
   public SocketImpl()
   {
      // TODO implement more constructor initializations logic EX:
      setResourceType("SOCKET");
      setUnknown(false);
   }
   
   /**
    * Constructor with PREV-SIBLING and NEXT-SIBLING as attributes.
    * 
    * @param prevSibling
    *        The previously created Socket handle
    * @param nextSibling
    *        The following Socket handle
    */
   public SocketImpl(handle prevSibling, handle nextSibling)
   {
      this();
      setPrevSibling(prevSibling);
      setNextSibling(nextSibling);
   }
   
   /** Constructor with socket parameter that is called from SERVER-SOCKET when a connection
    * is instantiated.
    * 
    * @param socket
    *        The socket on which this will be instantiated.
    */
   public SocketImpl(java.net.Socket socket)
   {
      this();
      sock = socket;
      setInputOutput();
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      // TODO research and  implement logic for it
      return true;
   }

   /**
    * Java implementation for getting the  LOCAL-HOST SOCKET attribute which 
    * contains the address in IP form of the machine to which the Socket is
    * connected.
    * 
    * @return The address in IP form of the machine to which the Socket is
    *         connected.
    */
   public character getLocalHost()
   {
      // TODO make correct implementation
      //      if (address != null)
      //      {
      //         return new character(address.getHostName());
      //      }
      //      else
      //      {
      //         return new character();
      //      }
      return new character();
   }
   
   /**
    * Java implementation for getting the  LOCAL-PORT SOCKET attribute which 
    * contains the local port of the machine to which the Socket is
    * connected.
    * 
    * @return The local port of the machine to which the Socket is
    */
   public integer getLocalPort()
   {
      // TODO Make correct implementation
   //      if (address != null)
   //      {
   //         return new integer(address.getPort());
   //      }
   //      else
   //      {
   //         return new integer();
   //      }
      return new integer();
   }
  
   /**
    * Java implementation for getting the  LOCAL-HOST SOCKET attribute which 
    * contains the address in IP form of the machine to which the Socket is
    * connected.
    * 
    * @return The remote address in IP form of the machine to which the Socket
    *         is connected.
    */
   public character getRemoteHost()
   {
      if (address != null)
      {
         return new character(address.getHostName());
      }
      else
      {
         return new character();
      }
   }
  
   /**
    * Java implementation for getting the  REMOTE-PORT SOCKET attribute which 
    * contains the local port of the machine to which the Socket is
    * connected.
    * 
    * @return The remote port of the machine to which the Socket is connected.
    */
   public integer getRemotePort()
   {
      if (address != null)
      {
         return new integer(address.getPort());
      }
      else
      {
         return new integer();
      }
   }
  
   /**
    * Java implementation for getting the  BYTES-READ SOCKET attribute which
    * contains the number of bytes the last read() method call returned from 
    * the socket.
    * 
    * @return The number of bytes read the read() method has returned from 
    * the socket.
    */
   public integer getBytesRead()
   {
      return new integer(bytesRead);
   }
  
   /**
    * Java implementation for getting the  BYTES-READ SOCKET attribute which
    * contains the number of bytes the last write() method has put in the
    * socket.
    * 
    * @return The number of bytes the last write() method has put in the
    *         socket.
    */
   public integer getBytesWritten()
   {
      return new integer(bytesWritten);
   }

   /**
    * Java implementation for getting the  SSL-SERVER-NAME SOCKET attribute 
    * which contains the SSL server name if the connection is made as SSL.
    * 
    * @return The SSL server name, if the connection is made as SSL or unknown
    *         otherwise
    */
   public character getSslServerName()
   {
      if (sock instanceof SSLSocket)
      {
         //implement it
      }
      return new character();
   }
  
   /**
    * Java implementation for the CONNECT() SOCKET method, this will attempt 
    * to connect to the host and port specified by the option argument, and 
    * returns <code>true</code> if connection succeeded, or <code>false</code>
    * otherwise.
    * <p>
    * The following options can be specified as valid connect() options:
    * <ul>
    *    <li><b>-H host-address</b> The host to which to connect, if not 
    *        specified localhost is used.
    *    <li><b>-S host-port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-nosessionreuse</b> If specified SSL session ID will not be 
    *        reused when reconnecting (ssl connection only).
    *    <li><b>-nohostverify</b> If specified no host verification will be 
    *        performed (ssl connection only).
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    * </ul>
    *  
    * @param  options
    *         The set of parameters for socket connection.
    * 
    * @return <code>true</code> if connection succeeded, or <code>false</code>
    *         otherwise
    */
   public logical connect(character options)
   {
      character errMsg = new character();
      integer errNo = new integer();
      character host = new character("localhost");
      integer port = new integer();
      logical ssl = new logical(false);
      logical noSessionReuse = new logical(false);
      logical noHostVerify = new logical(false);
      
      if (options.isUnknown())
      {
         ErrorManager.recordOrThrowError(5488, "The -S parameter must be specified to connect");
         return new logical(false);
      }
      
      getParams(options, errMsg, errNo, host, port, ssl, noSessionReuse, noHostVerify);
      
      if (!errMsg.isUnknown())
      {
         ErrorManager.recordOrThrowError(errNo.intValue(), errMsg.toStringMessage());
         ErrorManager.recordOrThrowError(5509, "Unable to process parameters.");
         ErrorManager.recordOrThrowError(5510, "Invalid parameter string.");
         return new logical(false);
      }
      
      InetSocketAddress address = new InetSocketAddress(host.toStringMessage(), port.intValue());
      
      if (ssl.booleanValue())
      {
         SSLSocketFactory factory = (SSLSocketFactory) SSLSocketFactory.getDefault();
         
         try
         {
            SSLSocket sslsock = (SSLSocket) factory.createSocket();
            sslsock.setReuseAddress(noSessionReuse.booleanValue());
            sslsock.connect(address);
            sock = sslsock;
         }
         catch (UnknownHostException e)
         {
            String msg = "Connection failure for host " + getLocalHost() +
                         " port " + port + " transport TCP.";
            ErrorManager.recordOrThrowError(9407, msg);
            return new logical(false);
         }
         catch (IOException e)
         {
            String msg = "Connection failure for host " + getLocalHost() +
                         " port " + port + " transport TCP.";
            ErrorManager.recordOrThrowError(9407, msg);
            return new logical(false);
         }
      }
      else
      {
         sock = new java.net.Socket();
      }
      
      setInputOutput();
      
      try
      {
         sock.connect(address);
         connected = true;
      }
      catch (IOException e)
      {
         String msg = "Connection failure for host " + getLocalHost() +
                      " port " + port + " transport TCP.";
         ErrorManager.recordOrThrowError(9407, msg);
         return new logical(false);
      }
      return new logical(true);
   }
   
   /**
    * Java implementation for the CONNECT() SOCKET method, this will attempt 
    * to connect to the host and port specified by the option argument, and 
    * returns <code>true</code> if connection succeeded, or <code>false</code>
    * otherwise.
    * <p>
    * The following options can be specified as valid connect() options:
    * <ul>
    *    <li><b>-H host-address</b> The host to which to connect, if not 
    *        specified localhost is used.
    *    <li><b>-S host-port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-nosessionreuse</b> If specified SSL session ID will not be 
    *        reused when reconnecting (ssl connection only).
    *    <li><b>-nohostverify</b> If specified no host verification will be 
    *        performed (ssl connection only).
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    * </ul>
    *  
    * @param  options
    *         The set of parameters for socket connection.
    * 
    * @return <code>true</code> if connection succeeded, or <code>false</code>
    *         otherwise
    */
   public logical connect(String options)
   {
      return connect(new character(options));
   }
  
   /**
    * Java implementation for the CONNECTED() SOCKET method, this will 
    * return <code>true</code> if the socket is connected or <code>false</code> 
    * otherwise.
    * 
    * @return <code>true</code> if the socket is connected or <code>false</code> 
    *         otherwise.
    */
   public logical connected()
   {
      return new logical(connected);
   }
  
   /**
    * Java implementation for the DISCONECT() SOCKET method, this will attempt 
    * to disconnect the socket and returns <code>true</code> if succeeded or 
    * <code>false</code> otherwise.
    * 
    * @return <code>true</code> if disconnect succeeded or <code>false</code> 
    *         otherwise.
    */
   public logical disconnect()
   {
      try
      {
         sock.close();
         connected = false;
         return new logical(true);
      }
      catch (IOException e)
      {
         return new logical(false);
      }

   }
  
   /**
    * Java implementation for the DISCONECT() SOCKET method, this will 
    * return the specified socket option value.
    * <p>
    * The following options names exist:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to retrieve.
    *         
    * @return The specified socket option value.
    */
   public character getSocketOption(character optionName)
   {
      // TODO Implement the actual logic
      return new character();
   }
  
   /**
    * Java implementation for the DISCONECT() SOCKET method, this will 
    * return the specified socket option value.
    *     * <p>
    * The following options names exist:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to retrieve.
    *         
    * @return The specified socket option value.
    */
   public character getSocketOption(String optionName)
   {
      return getSocketOption(new character(optionName));
   }
  
   /**
    * Java implementation for the GET-BYTES-AVAILABLE() SOCKET method, this 
    * will return the number of bytes that can be read at the current moment 
    * from the socket.
    * 
    * @return The number of bytes that can be read at the current moment 
    *         from the socket.
    */
   public integer getBytesAvailable()
   {
      
      return new integer(available());
   }
   
   /**
    * Java implementation for the GET-BYTES-AVAILABLE() SOCKET method, this 
    * will return the number of bytes that can be read at the current moment 
    * from the socket.
    * 
    * @return The number of bytes that can be read at the current moment 
    *         from the socket.
    */
   private int available()
   {
      int bytesAvail = 0;
      try
      {
         bytesAvail = in.available();
      }
      catch (IOException e)
      {
         // TODO Manage error here
      }
      return bytesAvail;
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, double pos, double len)
   {
      return read(buffer, new integer(pos), new integer(len));
   } 
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, NumberType pos, double len)
   {
     return read(buffer, pos, new integer(len));
   } 
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, double pos, NumberType len)
   {
      return read(buffer, new integer(pos), len);
   } 
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, NumberType pos, NumberType len)
   {
      // the default mode is READ-EXACT-NUM (2)
      return read(buffer, pos, len, Socket.READ_EXACT_NUM);
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, 
                       double     pos, 
                       double     len, 
                       double     mode)
   {
      return read(buffer, new integer(pos), new integer(len), new integer(mode));
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, 
                       NumberType pos, 
                       double     len, 
                       double     mode)
   {
     return read(buffer, new integer(pos), new integer(len), new integer(mode));
   } 
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer,
                       double     pos, 
                       NumberType len, 
                       double      mode)
   {
      return read(buffer, new integer(pos), len, new integer(mode));
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, 
                       NumberType pos, 
                       NumberType len, 
                       double     mode)
   {
     return read(buffer, pos, len, new integer(mode));
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer,
                       double     pos, 
                       double     len, 
                       NumberType mode)
   {
      return read(buffer, new integer(pos), new integer(len), mode);
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, 
                       NumberType pos, 
                       double     len, 
                       NumberType mode)
   {
      return read(buffer, pos, new integer(len), mode);
   }
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer,
                       double     pos, 
                       NumberType len, 
                       NumberType mode)
   {
      return read(buffer, new integer(pos), len, mode);
   } 
  
   /**
    * Java implementation for the READ() SOCKET method, this will read the 
    * specified number of bytes from the available data on the socket,
    * and put it inside the memory buffer argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param  buffer
    *         The memory buffer where read() puts the data.
    * @param  pos
    *         The position in the memory buffer where the data is put.
    * @param  len
    *         The number of bytes to read from the socket
    * @param  mode
    *         The mode to set for the reading operation.
    *        
    * @return <code>true</code> if the read was successful or <code>false</code> 
    *         otherwise.
    */
   public logical read(BinaryData buffer, 
                       NumberType pos, 
                       NumberType len, 
                       NumberType mode)
   {
      // TODO Add error handle for unknown parameters
      // reset bytesRead
      bytesRead = 0;
      try
      {
         if (mode.intValue() == Socket.READ_EXACT_NUM)
         {
            byte[] b = new byte[len.intValue()];
            int i = 0;
            int read = in.read();
            
            while (i < len.intValue() && (read != -1))
            {
               // TODO check possible case for read = -1 - end of input
               b[i] = (byte) read;
               i++;
               read = in.read();
            }
            bytesRead = len.intValue();
         }
         else if (mode.intValue() == Socket.READ_AVAILABLE)
         {
            
            int bAvail = available();
            
            while(bAvail == 0)
            {
               bAvail = available();
            }
            
            if (bAvail <= len.intValue())
            {
               byte[] b = new byte[available()];
               bytesRead = in.read();
               buffer.setByteRange(b, pos.intValue(), bAvail);
            }
            else
            {
               // TODO analyze more what can happen here
               byte[] b = new byte[len.intValue()];
               int i = 0;
               int read = in.read();
               
               while (i < len.intValue())
               {
                  b[i] = (byte) read;
                  i++;
                  read = in.read();
                  buffer.setByteRange(b, pos.intValue(), len.intValue());
               }
            }
         }
         else
         {
            //TODO ERROR handle this
         }
         
      }
      catch (IOException e)
      {
         // TODO Add error handling ;
         return new logical(false);
      }
      return new logical(true);
   }
  
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data
    * from the specified buffer location (position and number of bytes) to the
    * socket. 
    * 
    * @param  buffer
    *         The memory buffer from which the data will be written to the
    *         socket.
    * @param  pos
    *         The position in the memory buffer from where the data to be taken
    *         starts.
    * @param  len
    *         The number of bytes to take from the memory buffer.
    *         
    * @return <code>true</code> if the write was successful or <code>false</code> 
    *         otherwise.
    */
   public logical write(BinaryData buffer, double pos, double len)
   {
      return  write(buffer, new integer(pos), new integer(len));
   }
  
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data
    * from the specified buffer location (position and number of bytes) to the
    * socket. 
    * 
    * @param  buffer
    *         The memory buffer from which the data will be written to the
    *         socket.
    * @param  pos
    *         The position in the memory buffer from where the data to be taken
    *         starts.
    * @param  len
    *         The number of bytes to take from the memory buffer.
    *         
    * @return <code>true</code> if the write was successful or <code>false</code> 
    *         otherwise.
    */
   public logical write(BinaryData buffer, NumberType pos, double len)
   {
      return write(buffer, new integer(pos), new integer(len));
   }
  
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data
    * from the specified buffer location (position and number of bytes) to the
    * socket. 
    * 
    * @param  buffer
    *         The memory buffer from which the data will be written to the
    *         socket.
    * @param  pos
    *         The position in the memory buffer from where the data to be taken
    *         starts.
    * @param  len
    *         The number of bytes to take from the memory buffer.
    *         
    * @return <code>true</code> if the write was successful or <code>false</code> 
    *         otherwise.
    */
   public logical write(BinaryData buffer, double pos, NumberType len)
   {
      return write(buffer, new integer(pos), len);
   }
  
   /**
    * Java implementation for the WRITE() SOCKET method, this will write data
    * from the specified buffer location (position and number of bytes) to the
    * socket. 
    * 
    * @param  buffer
    *         The memory buffer from which the data will be written to the
    *         socket.
    * @param  pos
    *         The position in the memory buffer from where the data to be taken
    *         starts.
    * @param  len
    *         The number of bytes to take from the memory buffer.
    *         
    * @return <code>true</code> if the write was successful or <code>false</code> 
    *         otherwise..
    */
   public logical write(BinaryData buffer, NumberType pos, NumberType len)
   {
      // TODO Add error handle for unknown parameters
      bytesWritten = len.intValue();
      try
      {
         out.write(buffer.getByteRange(pos.intValue(), len.intValue()));
      }
      catch (IOException e)
      {
         // TODO Add error here
      }
      return new logical(true);
   }
  
   /**
   * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method, 
   * this will set the procedure to execute when a READ-RESPONSE event is 
   * received.
   * <p>
   * The set-read-response-procedure can also contain a handle to a higher 
   * level procedure which contains the read response procedure (to resolve 
   * any procedure context issues).
   * 
   * @param  procedureName
   *         The name of the procedure we want to set for read response.
   *         
   * @return <code>true</code> if the operation was successful or 
   *         <code>false</code> otherwise.
   */
   public logical setReadResponseProcedure(String procedureName)
   {
      return setReadResponseProcedure(new character(procedureName));
   }
  
   /**
   * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method, 
   * this will set the procedure to execute when a READ-RESPONSE event is 
   * received.
   * <p>
   * The set-read-response-procedure can also contain a handle to a higher 
   * level procedure which contains the read response procedure (to resolve 
   * any procedure context issues).
   * 
   * @param  procedureName
   *         The name of the procedure we want to set for read response.
   *         
   * @return <code>true</code> if the operation was successful or 
   *         <code>false</code> otherwise.
   */
   public logical setReadResponseProcedure(character procedureName)
   {
      // TODO Implement the actual logic
      return new logical(true);
   }

   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method, 
    * this will set the procedure to execute when a READ-RESPONSE event is 
    * received.
    * <p>
    * The set-read-response-procedure can also contain a handle to a higher
    * level procedure which contains the read response procedure (to resolve
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as read response.Optional
    *         
    * @return <code>true</code> if the operation was successful or
    *         <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(String procedureName, 
                                           handle procHandle)
   {
      return setReadResponseProcedure(new character(procedureName), procHandle);
   } 
   
   /**
    * Java implementation for the SET-READ-RESPONSE-PROCEDURE() SOCKET method, 
    * this will set the procedure to execute when a READ-RESPONSE event is 
    * received.
    * <p>
    * The set-read-response-procedure can also contain a handle to a higher
    * level procedure which contains the read response procedure (to resolve
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as read response.Optional
    *         
    * @return <code>true</code> if the operation was successful or
    *         <code>false</code> otherwise.
    */
   public logical setReadResponseProcedure(character procedureName, 
                                           handle procHandle)
   {
      // TODO Implement the actual logic
      return new logical(true);
   }
   /**
    * Java implementation for the SET-SOCKET-OPTION method, this will set the
    * value for the specified option name.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to set.
    * @param  optionVal
    *         The value of the option we want to set
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setSocketOption(String optionName, String optionVal)
   {
      return setSocketOption(new character(optionName), 
                             new character(optionVal));
   }
  
   /**
    * Java implementation for the SET-SOCKET-OPTION method, this will set the
    * value for the specified option name.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to set.
    * @param  optionVal
    *         The value of the option we want to set
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setSocketOption(character optionName, String optionVal)
   {
      return setSocketOption(optionName, new character(optionVal));
   }
  
   /**
    * Java implementation for the SET-SOCKET-OPTION method, this will set the
    * value for the specified option name.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to set.
    * @param  optionVal
    *         The value of the option we want to set
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setSocketOption(String optionName, character optionVal)
   {
      return setSocketOption(optionName, new character(optionVal));
   }
  
   /**
    * Java implementation for the SET-SOCKET-OPTION method, this will set the
    * value for the specified option name.
    * <p>
    * The following options are available:
    * <ul>
    *    <li>TCP-NODELAY
    *    <li>SO-LINGER
    *    <li>SO-KEEPALIVE
    *    <li>SO-REUSEADDR
    *    <li>SO-RCVBUF
    *    <li>SO-SNDBUF
    *    <li>SO-RCVTIMEO
    * <ul>
    * 
    * @param  optionName
    *         The name of the socket option we want to set.
    * @param  optionVal
    *         The value of the option we want to set
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setSocketOption(character optionName, character optionVal)
   {
      // TODO Implement the actual logic
      return new logical(true);
   }

   /**
    * Delete the resource.
    */
   @Override
   protected void resourceDelete()
   {
      // TODO Auto-generated method stub
   }
   
   /** Sets the input and output streams from the current socket*/
   private void setInputOutput()
   {
      try
      {
         out = new BufferedOutputStream(sock.getOutputStream());
         in = new BufferedInputStream(new BufferedInputStream(sock.getInputStream()));
      }
      catch (IOException e)
      {
         // TODO Manage error here
      }
      
   }
   /**
    * Utility method for parsing the connect parameters
    * 
    * @param options
    *        The string to parse.
    * @param errMsg
    *        The error message to set.
    * @param errNo
    *        The error number to set.
    * @param host
    *        The host which will be parsed
    * @param port
    *        The port which will be parse
    * @param ssl
    *        True if -ssl parameter will be parsed
    * @param noSessionReuse
    *        The -nosessionreuse parameter that will be parsed
    * @param noHostVerify
    *        The -nohostverify parameter that will be parsed.
    */
   private void getParams(character options,
                          character errMsg,
                          integer errNo,
                          character host,
                          integer port,
                          logical ssl,
                          logical noSessionReuse,
                          logical noHostVerify)
   {
    
      String strParams = options.toStringMessage();
      
      if (!strParams.toUpperCase().contains("-S") && !strParams.toUpperCase().contains("-PF"))
      {
         errMsg.assign("The -S parameter must be specified to connect");
         errNo.assign(5488);
      }
      String[] tokens = strParams.split(" ");
      
      for (int i = 0; i < tokens.length; i++)
      {
         if (!tokens[i].startsWith("-"))
         {
            errMsg.assign("Could not recognize arguments: " + tokens[i]);
            errNo.assign(301);
         }
         else
         {
            // if the option is the last element error
            if (i + 1 > tokens.length)
            {
               // Progress 4GL doesn't check if the argument is a valid accepted argument
               // this is just replicating the behavior
               errMsg.assign("You have not supplied a parameter for argument: " + tokens[i]);
               errNo.assign(1403);
               break;
            }
            if (tokens[i].equalsIgnoreCase("-H"))
            {
               host.assign(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-S"))
            {
               try
               {
                  port.assign(Integer.parseInt(tokens[i+1]));
                  i++;
               }
               catch(NumberFormatException e)
               {
                  errMsg.assign("Unknown service " + tokens[i+1] + " transport TCP");
                  errNo.assign(5484);
                  break;
               }               
            }
            else if (tokens[i].equalsIgnoreCase("-ssl"))
            {
               ssl.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-nosessionreuse"))
            {
               noSessionReuse.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-nohostverify"))
            {
               noHostVerify.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-pf"))
            {
               File configFile = new File(tokens[i+1]);
               if (!configFile.exists() && configFile.isFile())
               {
                  errMsg.assign("Unable to open parameter file " + tokens[i+1] + ", errno 2.");
                  errNo.assign(1247);
                  break;
               }
               else
               {
                  // clear any existing errors
                  errMsg = new character();
                  BufferedReader inputReader = null;
                  
                  try
                  {
                     inputReader = new BufferedReader(new FileReader(configFile));
                  }
                  catch (FileNotFoundException e)
                  {
                    
                  }
                  
                  try
                  {
                     StringBuffer b =  new StringBuffer();
                     int asc = inputReader.read();
                     while (asc != -1)
                     {
                        char chr = (char) asc;
                        b.append(chr);
                        asc = inputReader.read();
                     }
                     character fileOpts = new character(b.toString());
                     getParams(fileOpts, errMsg, errNo, host, 
                               port, ssl, noSessionReuse, noHostVerify);
                  }
                  catch (IOException e)
                  {
                     errMsg.assign("An error has occured while reading from the parameter file " +
                                   tokens[i+1] + ", errno 2.");
                  }
                  
               }
            }
         }
      }
   }
}
